use egui::emath::RectTransform;
use egui::*;
use flare_shader::Coefs;
use std::ops::Add;

/// Radius (in pixels) of the transform element draw circle
const ELEMENT_DRAW_RADIUS_PX: f32 = 7.0;

/// Stroke size (in pixels) of the transform element draw circle
const ELEMENT_DRAW_STROKE_PX: f32 = 2.0;

/// Offset (in pixels) of the transform element draw circle
const ELEMENT_DRAW_OFFSET_PX: Vec2 = vec2(-2.0, -2.0);

fn test_in_circle(pt: Pos2, center: Pos2, radius: f32) -> bool {
    ((pt.x - center.x).powf(2.0) + (pt.y - center.y).powf(2.0)) <= radius.powf(2.0)
}

fn test_in_triangle(pt: Pos2, v1: Pos2, v2: Pos2, v3: Pos2) -> bool {
    // https://stackoverflow.com/a/2049593
    let sign = |p1: Pos2, p2: Pos2, p3: Pos2| -> f32 {
        (p1.x - p3.x) * (p2.y - p3.y) - (p2.x - p3.x) * (p1.y - p3.y)
    };

    let d1 = sign(pt, v1, v2);
    let d2 = sign(pt, v2, v3);
    let d3 = sign(pt, v3, v1);

    let has_neg = [d1, d2, d3].iter().any(|v| *v < 0.0);
    let has_pos = [d1, d2, d3].iter().any(|v| *v > 0.0);

    !(has_neg && has_pos)
}

#[derive(Copy, Clone, Debug, PartialEq)]
pub enum TransformElement {
    Origin,
    X,
    Y,
}

/// Affine coefficients expressed as three points of a triangle
#[derive(Copy, Clone, Debug, PartialEq)]
pub struct TransformTriangle {
    origin: Pos2,
    x: Pos2,
    y: Pos2,
}

impl TransformTriangle {
    pub fn new(origin: Pos2, x: Pos2, y: Pos2) -> Self {
        Self { origin, x, y }
    }

    pub fn interact_drag(self, element: TransformElement, drag_delta: Vec2) -> Self {
        match element {
            TransformElement::X => Self::new(self.origin, self.x + drag_delta, self.y),
            TransformElement::Y => Self::new(self.origin, self.x, self.y + drag_delta),
            TransformElement::Origin => Self::new(
                self.origin + drag_delta,
                self.x + drag_delta,
                self.y + drag_delta,
            ),
        }
    }

    pub fn transform_pos(self, rect: RectTransform) -> Self {
        Self {
            origin: rect.transform_pos(self.origin),
            x: rect.transform_pos(self.x),
            y: rect.transform_pos(self.y),
        }
    }

    pub fn is_hovered(
        &self,
        hover_pos: Pos2,
        ifs_to_screen: RectTransform,
    ) -> Option<TransformElement> {
        let origin_pos = ifs_to_screen.transform_pos(self.origin);
        let x_pos = ifs_to_screen.transform_pos(self.x);
        let y_pos = ifs_to_screen.transform_pos(self.y);

        if test_in_circle(hover_pos, x_pos, ELEMENT_DRAW_RADIUS_PX) {
            Some(TransformElement::X)
        } else if test_in_circle(hover_pos, y_pos, ELEMENT_DRAW_RADIUS_PX) {
            Some(TransformElement::Y)
        } else if test_in_circle(hover_pos, origin_pos, ELEMENT_DRAW_RADIUS_PX)
            || test_in_triangle(hover_pos, origin_pos, x_pos, y_pos)
        {
            Some(TransformElement::Origin)
        } else {
            None
        }
    }
}

impl From<Coefs> for TransformTriangle {
    fn from(value: Coefs) -> Self {
        let origin = pos2(value.c, -value.f);
        Self {
            origin,
            x: origin + vec2(value.a, -value.d),
            y: origin + vec2(-value.b, value.e),
        }
    }
}

impl Into<Coefs> for TransformTriangle {
    fn into(self) -> Coefs {
        Coefs {
            a: self.x.x - self.origin.x,
            b: self.origin.x - self.y.x,
            c: self.origin.x,
            d: self.origin.y - self.x.y,
            e: self.y.y - self.origin.y,
            f: -self.origin.y,
        }
    }
}

impl Add<Vec2> for TransformTriangle {
    type Output = TransformTriangle;

    fn add(self, rhs: Vec2) -> Self::Output {
        Self {
            origin: self.origin + rhs,
            x: self.x + rhs,
            y: self.y + rhs,
        }
    }
}

/// Widget for manipulating IFS transform affine coefficients
#[derive(Copy, Clone, Debug)]
pub struct TransformEditor {
    /// Center point (in IFS coordinates) of the editor window
    center_ifs: Pos2,

    /// Total range (in IFS coordinates) of the editor window
    range_ifs: f32,

    /// Hover position (in screen coordinates) of the cursor on the previous update.
    ///
    /// Because of input latency during large drag motions, `egui`'s drag motion
    /// isn't precise enough to update transform coefficients. Instead, track the
    /// cursor position directly
    hover_pos: Option<Pos2>,

    /// Transform index the cursor is hovering over
    hover_index: Option<usize>,

    /// Specific element of the transform hovered by the cursor
    hover_element: Option<TransformElement>,

    /// Transform index the cursor is dragging
    drag_index: Option<usize>,
}

impl Default for TransformEditor {
    fn default() -> Self {
        Self {
            center_ifs: Pos2::ZERO,
            range_ifs: 4.0,
            hover_pos: None,
            hover_index: None,
            hover_element: None,
            drag_index: None,
        }
    }
}

fn build_viewport_ifs(aspect_ratio: f32, center_ifs: Pos2, range_ifs: f32) -> Rect {
    let size_ifs = if aspect_ratio >= 1.0 {
        vec2(range_ifs * aspect_ratio, range_ifs)
    } else {
        vec2(range_ifs, range_ifs / aspect_ratio)
    };
    let min_ifs = center_ifs - size_ifs / 2.0;
    let max_ifs = center_ifs + size_ifs / 2.0;

    // IFS coordinates follow the "value increases from top left to bottom right" convention.
    // Because we want coordinates to behave like a Cartesian plot, the Y-axis is flipped
    Rect::from_min_max(pos2(min_ifs.x, max_ifs.y), pos2(max_ifs.x, min_ifs.y))
}

impl TransformEditor {
    /// Interact with the provided transform coefficients. Returns the index of the focused transform, if any
    pub fn interact(&mut self, ui: &mut Ui, transforms: &mut [Coefs]) -> Option<usize> {
        let (response, painter) = ui.allocate_painter(ui.available_size(), Sense::drag());

        if transforms.is_empty() {
            return None;
        }

        let interact_rect = response.interact_rect;
        let ifs_rect = build_viewport_ifs(
            interact_rect.aspect_ratio(),
            self.center_ifs,
            self.range_ifs,
        );

        // Update internal state based on screen interactions, then paint to screen
        let ifs_to_screen = RectTransform::from_to(ifs_rect, interact_rect);
        self.interact_update(
            ui.input(|i| i.pointer.interact_pos()),
            ui.input(|i| i.pointer.primary_pressed()),
            ui.input(|i| i.pointer.primary_released()),
            ifs_to_screen,
            transforms,
        );

        self.interact_draw(painter, ifs_to_screen, transforms);

        self.drag_index
    }

    /// Update state of the provided transform coefficients based on current interactions,
    /// return the transform coefficients that have claimed focus (if any)
    fn interact_update(
        &mut self,
        hover_pos: Option<Pos2>,
        primary_pressed: bool,
        primary_released: bool,
        ifs_to_screen: RectTransform,
        transforms: &mut [Coefs],
    ) {
        // If the cursor is not in this widget, reset state
        if hover_pos.is_none() {
            self.hover_pos = None;
            self.hover_index = None;
            self.hover_element = None;
            self.drag_index = None;
            return;
        }

        let hover_pos = hover_pos.unwrap();

        // If the transform array was modified, reset state and then proceed
        if self.hover_index.map_or(false, |i| i >= transforms.len())
            || self.drag_index.map_or(false, |i| i >= transforms.len())
        {
            self.hover_index = None;
            self.hover_element = None;
            self.drag_index = None;
        }

        // If a transform is being dragged, update its position
        if self.drag_index.is_some() {
            let last_hover_pos = self.hover_pos.unwrap();
            let hover_index = self.hover_index.unwrap();
            let hover_element = self.hover_element.unwrap();

            let drag_delta_ifs = (hover_pos - last_hover_pos) / ifs_to_screen.scale();
            let transform_triangle: TransformTriangle = transforms[hover_index].into();
            transforms[hover_index] = transform_triangle
                .interact_drag(hover_element, drag_delta_ifs)
                .into();
        }

        // Store the hover pos for use on the next update
        self.hover_pos = Some(hover_pos);

        // Check if the currently-hovered transform is still hovered
        let mut hover_found = false;

        if let Some(hover_index) = self.hover_index {
            let transform_triangle: TransformTriangle = transforms[hover_index].into();
            self.hover_element = transform_triangle.is_hovered(hover_pos, ifs_to_screen);
            hover_found = self.hover_element.is_some();
        }

        // Check if any transform is hovered
        if !hover_found {
            for (i, transform) in transforms.iter().enumerate() {
                let transform_triangle: TransformTriangle = (*transform).into();
                self.hover_element = transform_triangle.is_hovered(hover_pos, ifs_to_screen);

                if self.hover_element.is_some() {
                    hover_found = true;
                    self.hover_index = Some(i);
                    break;
                }
            }
        }

        if !hover_found {
            // No hovers found, reset state
            self.hover_index = None;
            self.hover_element = None;
            self.drag_index = None;
        }

        // Check drag state
        if primary_pressed && self.hover_index.is_some() {
            self.drag_index = self.hover_index;
        }
        if primary_released {
            self.drag_index = None;
        }
    }

    fn interact_draw_transform(
        hover_element: Option<TransformElement>,
        painter: &Painter,
        ifs_to_screen: RectTransform,
        transform: Coefs,
    ) {
        // `epaint` doesn't provide an option for whether the stroke is drawn inside, in the middle,
        // or outside the shape to paint. In manual testing, hover detection works best when assuming
        // the stroke is outside the shape.
        // Also in manual testing, hover detection seems to work best when drawing the circle at
        // a slight offset to the actual center position. Not clear why.
        let transform_triangle: TransformTriangle = transform.into();
        let draw_triangle =
            transform_triangle.transform_pos(ifs_to_screen) + ELEMENT_DRAW_OFFSET_PX;

        let stroke = Stroke::new(ELEMENT_DRAW_STROKE_PX, Color32::BLUE);
        painter.circle_stroke(
            draw_triangle.origin,
            ELEMENT_DRAW_RADIUS_PX - ELEMENT_DRAW_STROKE_PX,
            stroke,
        );
        painter.circle_stroke(
            draw_triangle.x,
            ELEMENT_DRAW_RADIUS_PX - ELEMENT_DRAW_STROKE_PX,
            stroke,
        );
        painter.circle_stroke(
            draw_triangle.y,
            ELEMENT_DRAW_RADIUS_PX - ELEMENT_DRAW_STROKE_PX,
            stroke,
        );

        let body_alpha: u8 = if hover_element.is_some() { 8 } else { 0 };
        let body_fill = Color32::from_rgba_unmultiplied(0, 0, u8::MAX, body_alpha);
        let body = Shape::convex_polygon(
            vec![draw_triangle.origin, draw_triangle.x, draw_triangle.y],
            body_fill,
            stroke,
        );
        painter.add(body);
    }

    /// Draw the provided transform coefficients to the screen.
    fn interact_draw(&self, painter: Painter, ifs_to_screen: RectTransform, transforms: &[Coefs]) {
        transforms
            .iter()
            .enumerate()
            .filter(|(i, _)| Some(*i) != self.hover_index)
            .for_each(|(_, transform)| {
                Self::interact_draw_transform(None, &painter, ifs_to_screen, *transform)
            });

        self.hover_index.map(|i| {
            Self::interact_draw_transform(
                self.hover_element,
                &painter,
                ifs_to_screen,
                transforms[i],
            )
        });
    }

    pub fn interact_debug(&self, ui: &mut Ui) {
        Grid::new("transform_editor_debug")
            .num_columns(3)
            .show(ui, |ui| {
                ui.set_width(ui.available_width());
                ui.label(format!("Hover index: {:?}", self.hover_index));
                ui.label(format!("Hover element: {:?}", self.hover_element));
                ui.label(format!("Drag index: {:?}", self.drag_index));
            });
    }
}
