extern crate core;

use eframe::Frame;
use eframe::epaint::PaintCallbackInfo;
use egui::{Context, Rect};
use egui_wgpu::{CallbackResources, CallbackTrait, ScreenDescriptor};
use flare_shader::{
    AccumConstants, CameraConstants, Coefs, ThreadState, Transform, Variation, VariationKind,
    ViewportConstants,
};
use futures::channel::oneshot;
use futures_executor::block_on;
use image::codecs::png::PngEncoder;
use image::{ExtendedColorType, ImageEncoder};
use rand::thread_rng;
use std::fs::File;
use wgpu::util::DeviceExt;
use wgpu::{CommandBuffer, CommandEncoder, Device, Queue, RenderPass};

pub mod transform_editor;

struct RenderGroup {
    device: wgpu::Device,
    queue: wgpu::Queue,
    bind_group_layout: wgpu::BindGroupLayout,
    bind_group: wgpu::BindGroup,
    viewport_constants_buffer: wgpu::Buffer,
    accum_constants: AccumConstants,
    accum_constants_buffer: wgpu::Buffer,
    camera_constants_buffer: wgpu::Buffer,
    image_accum_buffer: wgpu::Buffer,
}

impl RenderGroup {
    pub fn bind_group_layout(device: &wgpu::Device) -> wgpu::BindGroupLayout {
        device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
            label: Some("render"),
            entries: &[
                // image_constants
                wgpu::BindGroupLayoutEntry {
                    binding: 0,
                    visibility: wgpu::ShaderStages::COMPUTE | wgpu::ShaderStages::FRAGMENT,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: true },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
                // accum_constants
                wgpu::BindGroupLayoutEntry {
                    binding: 1,
                    visibility: wgpu::ShaderStages::COMPUTE | wgpu::ShaderStages::FRAGMENT,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: true },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
                // camera_constants
                wgpu::BindGroupLayoutEntry {
                    binding: 2,
                    visibility: wgpu::ShaderStages::COMPUTE | wgpu::ShaderStages::FRAGMENT,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: true },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
                // accum_image
                wgpu::BindGroupLayoutEntry {
                    binding: 3,
                    visibility: wgpu::ShaderStages::COMPUTE | wgpu::ShaderStages::FRAGMENT,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: false },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
            ],
        })
    }

    fn build_image_accum_buffer(device: &wgpu::Device, width: i32, height: i32) -> wgpu::Buffer {
        let pixel_count = (width * height) as u64;
        let size = pixel_count * 4 * size_of::<f32>() as u64;
        device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("accum_image"),
            size,
            usage: wgpu::BufferUsages::STORAGE | wgpu::BufferUsages::COPY_SRC,
            mapped_at_creation: false,
        })
    }

    fn build_bind_group(
        device: &wgpu::Device,
        layout: &wgpu::BindGroupLayout,
        image_constants_buffer: &wgpu::Buffer,
        accum_constants_buffer: &wgpu::Buffer,
        camera_constants_buffer: &wgpu::Buffer,
        image_accum_buffer: &wgpu::Buffer,
    ) -> wgpu::BindGroup {
        device.create_bind_group(&wgpu::BindGroupDescriptor {
            label: Some("render"),
            layout,
            entries: &[
                wgpu::BindGroupEntry {
                    binding: 0,
                    resource: image_constants_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 1,
                    resource: accum_constants_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 2,
                    resource: camera_constants_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 3,
                    resource: image_accum_buffer.as_entire_binding(),
                },
            ],
        })
    }

    pub fn new(device: &wgpu::Device, queue: &wgpu::Queue, width: i32, height: i32) -> Self {
        let bind_group_layout = Self::bind_group_layout(device);

        let viewport_constants_buffer = device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("viewport_constants"),
            size: size_of::<ViewportConstants>() as u64,
            usage: wgpu::BufferUsages::STORAGE | wgpu::BufferUsages::COPY_DST,
            mapped_at_creation: false,
        });

        let accum_constants = AccumConstants::new(width, height);
        let accum_constants_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("accum_constants"),
            contents: bytemuck::cast_slice(&[accum_constants.clone()]),
            usage: wgpu::BufferUsages::STORAGE | wgpu::BufferUsages::COPY_DST,
        });

        let min_dimension = width.min(height);
        let initial_camera = CameraConstants::new(min_dimension as f32);
        let camera_constants_buffer =
            device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
                label: Some("camera_constants"),
                contents: bytemuck::cast_slice(&[initial_camera]),
                usage: wgpu::BufferUsages::STORAGE | wgpu::BufferUsages::COPY_DST,
            });

        let image_accum_buffer = Self::build_image_accum_buffer(device, width, height);

        let bind_group = Self::build_bind_group(
            device,
            &bind_group_layout,
            &viewport_constants_buffer,
            &accum_constants_buffer,
            &camera_constants_buffer,
            &image_accum_buffer,
        );

        Self {
            device: device.clone(),
            queue: queue.clone(),
            bind_group_layout,
            viewport_constants_buffer,
            accum_constants,
            accum_constants_buffer,
            camera_constants_buffer,
            image_accum_buffer,
            bind_group,
        }
    }

    pub fn set_accum_dimensions(&mut self, width: i32, height: i32) {
        self.accum_constants = AccumConstants::new(width, height);
        self.queue.write_buffer(
            &self.accum_constants_buffer,
            0,
            bytemuck::cast_slice(&[self.accum_constants]),
        );

        self.image_accum_buffer = Self::build_image_accum_buffer(&self.device, width, height);
        self.bind_group = Self::build_bind_group(
            &self.device,
            &self.bind_group_layout,
            &self.viewport_constants_buffer,
            &self.accum_constants_buffer,
            &self.camera_constants_buffer,
            &self.image_accum_buffer,
        );
    }

    pub fn set_viewport_dimensions(&mut self, rect: egui::Rect) {
        let offset = rect.left_top();
        let viewport_constants = ViewportConstants::new(
            offset.x as i32,
            offset.y as i32,
            rect.width() as i32,
            rect.height() as i32,
        );
        self.queue.write_buffer(
            &self.viewport_constants_buffer,
            0,
            bytemuck::cast_slice(&[viewport_constants]),
        )
    }

    pub fn bind_group(&self) -> &wgpu::BindGroup {
        &self.bind_group
    }
}

struct IfsGroup {
    device: wgpu::Device,
    bind_group_layout: wgpu::BindGroupLayout,
    thread_state_buffer: wgpu::Buffer,
    transforms_buffer: Option<wgpu::Buffer>,
    variations_buffer: Option<wgpu::Buffer>,
    bind_group: Option<wgpu::BindGroup>,
}

impl IfsGroup {
    pub fn bind_group_layout(device: &wgpu::Device) -> wgpu::BindGroupLayout {
        device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
            label: Some("accum"),
            entries: &[
                // thread_state
                wgpu::BindGroupLayoutEntry {
                    binding: 0,
                    visibility: wgpu::ShaderStages::COMPUTE,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: false },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
                // transforms
                wgpu::BindGroupLayoutEntry {
                    binding: 1,
                    visibility: wgpu::ShaderStages::COMPUTE,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: true },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
                // variations
                wgpu::BindGroupLayoutEntry {
                    binding: 2,
                    visibility: wgpu::ShaderStages::COMPUTE,
                    ty: wgpu::BindingType::Buffer {
                        ty: wgpu::BufferBindingType::Storage { read_only: true },
                        has_dynamic_offset: false,
                        min_binding_size: None,
                    },
                    count: None,
                },
            ],
        })
    }

    pub fn new(device: &wgpu::Device, queue: &wgpu::Queue) -> Self {
        let mut seed_rng = thread_rng();
        let mut thread_states = vec![ThreadState::new(&mut seed_rng)];
        while thread_states.len() < 64 {
            thread_states.push(thread_states[thread_states.len() - 1].jump());
        }

        let thread_state_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("thread_state"),
            contents: bytemuck::cast_slice(thread_states.as_slice()),
            usage: wgpu::BufferUsages::STORAGE,
        });

        Self {
            device: device.clone(),
            bind_group_layout: Self::bind_group_layout(device),
            thread_state_buffer,
            transforms_buffer: None,
            variations_buffer: None,
            bind_group: None,
        }
    }

    pub fn set_transforms(&mut self, transforms: &[Transform]) {
        let transforms_buffer = self
            .device
            .create_buffer_init(&wgpu::util::BufferInitDescriptor {
                label: Some("transforms"),
                contents: bytemuck::cast_slice(transforms),
                usage: wgpu::BufferUsages::STORAGE,
            });
        let _ = self.transforms_buffer.insert(transforms_buffer);
    }

    pub fn set_variations(&mut self, variations: &[Variation]) {
        let variations_buffer = self
            .device
            .create_buffer_init(&wgpu::util::BufferInitDescriptor {
                label: Some("variations"),
                contents: bytemuck::cast_slice(variations),
                usage: wgpu::BufferUsages::STORAGE,
            });
        let _ = self.variations_buffer.insert(variations_buffer);
    }

    pub fn bind_group(&mut self) -> &wgpu::BindGroup {
        if self.bind_group.is_none() {
            let transforms_buffer = self
                .transforms_buffer
                .as_ref()
                .expect("No transforms available");
            let variations_buffer = self
                .variations_buffer
                .as_ref()
                .expect("No variations available");

            let bind_group = self.device.create_bind_group(&wgpu::BindGroupDescriptor {
                label: Some("ifs"),
                layout: &self.bind_group_layout,
                entries: &[
                    wgpu::BindGroupEntry {
                        binding: 0,
                        resource: self.thread_state_buffer.as_entire_binding(),
                    },
                    wgpu::BindGroupEntry {
                        binding: 1,
                        resource: transforms_buffer.as_entire_binding(),
                    },
                    wgpu::BindGroupEntry {
                        binding: 2,
                        resource: variations_buffer.as_entire_binding(),
                    },
                ],
            });

            let _ = self.bind_group.insert(bind_group);
        }

        self.bind_group.as_ref().unwrap()
    }
}

struct AccumPipeline {
    pipeline: wgpu::ComputePipeline,
}

impl AccumPipeline {
    pub fn new(device: &wgpu::Device, module: &wgpu::ShaderModule) -> Self {
        let render_layout = RenderGroup::bind_group_layout(device);
        let ifs_layout = IfsGroup::bind_group_layout(device);

        let pipeline_layout = device.create_pipeline_layout(&wgpu::PipelineLayoutDescriptor {
            label: Some("accum"),
            bind_group_layouts: &[&render_layout, &ifs_layout],
            push_constant_ranges: &[],
        });

        let pipeline = device.create_compute_pipeline(&wgpu::ComputePipelineDescriptor {
            label: Some("accum"),
            layout: Some(&pipeline_layout),
            module,
            entry_point: Some("main_cs"),
            compilation_options: Default::default(),
            cache: None,
        });

        Self { pipeline }
    }
}

struct RenderPipeline {
    pipeline: wgpu::RenderPipeline,
}

impl RenderPipeline {
    pub fn new(
        device: &wgpu::Device,
        module: &wgpu::ShaderModule,
        target: &wgpu::TextureFormat,
    ) -> Self {
        let render_layout = RenderGroup::bind_group_layout(device);
        let pipeline_layout = device.create_pipeline_layout(&wgpu::PipelineLayoutDescriptor {
            label: Some("render"),
            bind_group_layouts: &[&render_layout],
            push_constant_ranges: &[],
        });
        let pipeline = device.create_render_pipeline(&wgpu::RenderPipelineDescriptor {
            label: Some("render"),
            layout: Some(&pipeline_layout),
            vertex: wgpu::VertexState {
                module,
                entry_point: Some("main_vs"),
                compilation_options: Default::default(),
                buffers: &[],
            },
            primitive: Default::default(),
            depth_stencil: None,
            multisample: Default::default(),
            fragment: Some(wgpu::FragmentState {
                module,
                entry_point: Some("main_fs"),
                compilation_options: Default::default(),
                targets: &[Some((*target).into())],
            }),
            multiview: None,
            cache: None,
        });
        Self { pipeline }
    }
}

struct FlareAssets {
    render_group: RenderGroup,
    ifs_group: IfsGroup,
    accum_pipeline: AccumPipeline,
    render_pipeline: RenderPipeline,
}

struct Flare {
    first_draw: bool,
}

impl Flare {
    pub fn new(cc: &eframe::CreationContext<'_>, width: i32, height: i32) -> Self {
        let cc_wgpu = cc
            .wgpu_render_state
            .as_ref()
            .expect("Unable to acquire wgpu render state");
        let device = cc_wgpu.device.clone();
        let queue = cc_wgpu.queue.clone();

        let module_spirv = wgpu::include_spirv!(concat!(env!("OUT_DIR"), "/flare.spv"));
        let module = device.create_shader_module(module_spirv);

        // Draw a gasket
        let mut ifs_group = IfsGroup::new(&device, &queue);

        let f0 = Transform::new(Coefs::new(0.5, 0., 0., 0., 0.5, 0.), 0, 1, 1.0);
        let f1 = Transform::new(Coefs::new(0.5, 0., 0.5, 0., 0.5, 0.), 0, 1, 1.0);
        let f2 = Transform::new(Coefs::new(0.5, 0., 0., 0., 0.5, 0.5), 0, 1, 1.0);
        let variation = Variation::new(VariationKind::Linear, 1.0);

        ifs_group.set_transforms(&[f0, f1, f2]);
        ifs_group.set_variations(&[variation]);

        cc_wgpu
            .renderer
            .write()
            .callback_resources
            .insert(FlareAssets {
                render_group: RenderGroup::new(&device, &queue, width, height),
                ifs_group,
                accum_pipeline: AccumPipeline::new(&device, &module),
                render_pipeline: RenderPipeline::new(&device, &module, &cc_wgpu.target_format),
            });

        Self { first_draw: true }
    }
}

impl eframe::App for Flare {
    fn update(&mut self, ctx: &Context, _frame: &mut Frame) {
        egui::TopBottomPanel::top("top_panel").show(ctx, |ui| {
            egui::menu::bar(ui, |ui| {
                ui.menu_button("File", |ui| {
                    if ui.button("Debug").clicked() {
                        ui.painter().add(egui_wgpu::Callback::new_paint_callback(
                            Rect::ZERO,
                            DebugIfs,
                        ));
                        ui.close_menu();
                    }
                    if ui.button("Quit").clicked() {
                        ctx.send_viewport_cmd(egui::ViewportCommand::Close);
                    }
                });
            });
        });

        egui::CentralPanel::default().show(ctx, |ui| {
            egui::Frame::canvas(ui.style()).show(ui, |ui| {
                let rect = ui.available_rect_before_wrap();
                let response = ui.allocate_rect(rect, egui::Sense::click());
                let painter = PaintIfs {
                    run_accumulate: self.first_draw || response.clicked(),
                    viewport: rect,
                };
                ui.painter()
                    .add(egui_wgpu::Callback::new_paint_callback(rect, painter));
            });
        });

        self.first_draw = false
    }
}

struct DebugIfs;

impl CallbackTrait for DebugIfs {
    fn prepare(
        &self,
        device: &Device,
        queue: &Queue,
        _screen_descriptor: &ScreenDescriptor,
        egui_encoder: &mut CommandEncoder,
        callback_resources: &mut CallbackResources,
    ) -> Vec<CommandBuffer> {
        let flare_assets = callback_resources
            .get::<FlareAssets>()
            .expect("Missing assets");

        let export_buffer = device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("image_export"),
            size: flare_assets.render_group.image_accum_buffer.size(),
            usage: wgpu::BufferUsages::COPY_DST | wgpu::BufferUsages::MAP_READ,
            mapped_at_creation: false,
        });

        let device_extra = device.clone();
        let width_extra = flare_assets.render_group.accum_constants.width as u32;
        let height_extra = flare_assets.render_group.accum_constants.height as u32;

        egui_encoder.copy_buffer_to_buffer(
            &flare_assets.render_group.image_accum_buffer,
            0,
            &export_buffer,
            0,
            export_buffer.size(),
        );
        queue.on_submitted_work_done(move || {
            let slice = export_buffer.slice(..);
            let (sender, receiver) = oneshot::channel();

            slice.map_async(wgpu::MapMode::Read, move |result| {
                if let Ok(result) = result {
                    sender.send(result).unwrap()
                }
            });

            device_extra.poll(wgpu::Maintain::Wait);
            block_on(receiver).expect("Unable to map buffer");

            let result: Vec<f32> = {
                let data = slice.get_mapped_range();
                bytemuck::cast_slice(&data).to_vec()
            };
            export_buffer.unmap();

            let result_u8: Vec<u8> = result.iter().map(|v| (*v * 256.0) as u8).collect();

            let file = File::create("accum_image.png").expect("Unable to create output image");
            let png_encoder = PngEncoder::new(file);

            png_encoder
                .write_image(
                    &result_u8,
                    width_extra,
                    height_extra,
                    ExtendedColorType::Rgba8,
                )
                .expect("Unable to write output");
        });

        vec![]
    }

    fn paint(
        &self,
        _info: PaintCallbackInfo,
        _render_pass: &mut RenderPass<'static>,
        _callback_resources: &CallbackResources,
    ) {
    }
}

struct PaintIfs {
    run_accumulate: bool,
    viewport: egui::Rect,
}

impl CallbackTrait for PaintIfs {
    fn prepare(
        &self,
        _device: &Device,
        _queue: &Queue,
        _screen_descriptor: &ScreenDescriptor,
        egui_encoder: &mut CommandEncoder,
        callback_resources: &mut CallbackResources,
    ) -> Vec<CommandBuffer> {
        let flare_assets = callback_resources
            .get_mut::<FlareAssets>()
            .expect("Missing assets");

        if self.run_accumulate {
            flare_assets
                .render_group
                .set_accum_dimensions(self.viewport.width() as i32, self.viewport.height() as i32);

            let mut pass = egui_encoder.begin_compute_pass(&wgpu::ComputePassDescriptor {
                label: Some("accum"),
                timestamp_writes: None,
            });
            pass.set_pipeline(&flare_assets.accum_pipeline.pipeline);
            pass.set_bind_group(0, flare_assets.render_group.bind_group(), &[]);
            pass.set_bind_group(1, flare_assets.ifs_group.bind_group(), &[]);
            pass.dispatch_workgroups(1, 1, 1);
        }

        flare_assets
            .render_group
            .set_viewport_dimensions(self.viewport);

        vec![]
    }

    fn paint(
        &self,
        _info: PaintCallbackInfo,
        render_pass: &mut RenderPass<'static>,
        callback_resources: &CallbackResources,
    ) {
        let flare_assets = callback_resources
            .get::<FlareAssets>()
            .expect("Missing assets");

        render_pass.set_pipeline(&flare_assets.render_pipeline.pipeline);
        render_pass.set_bind_group(0, flare_assets.render_group.bind_group(), &[]);
        render_pass.draw(0..3, 0..1);
    }
}

fn main() -> eframe::Result {
    env_logger::init();

    let initial_dimensions = egui::vec2(800., 600.);
    let native_options = eframe::NativeOptions {
        viewport: egui::ViewportBuilder::default().with_inner_size(initial_dimensions),
        renderer: eframe::Renderer::Wgpu,
        ..Default::default()
    };

    eframe::run_native(
        "flare",
        native_options,
        Box::new(|cc| {
            Ok(Box::new(Flare::new(
                cc,
                initial_dimensions.x as i32,
                initial_dimensions.y as i32,
            )))
        }),
    )
}
